import logging
from typing import Union, Tuple, List, Dict

import numpy as np
from datasets.metric import Metric
from torch import Tensor
from transformers import (
    AutoModelForTokenClassification,
    DataCollatorForTokenClassification,
)

from .transformers_base_wrapper import TransformersBaseWrapper
from .wrapper_encoder import WrapperEncoder

log = logging.getLogger()


class WrapperNer(WrapperEncoder, TransformersBaseWrapper):
    @staticmethod
    def align_labels_with_tokens(
        labels: List[int], word_ids: List[int], tag_for_subtokens=None
    ):
        """
        Function to adjust the labels according to the tokenized input
        :param labels: original list of tags
        :param word_ids: list with ids of the original words in the tokenized list of tokens;
        can be extracted via `inputs.word_ids()`, where `inputs = tokenizer(...)`
        :param tag_for_subtokens: how to tokenize subtokens (BPEs) of a word. Have three options:
        imagine we have a word `loving` -> [`lov`, `###ing`].
        Option 1 (default) = None: `lov` -> B-TAG, `###ing` -> O
        Option 2 = 'i': `lov` -> B-TAG, `###ing` -> I-TAG
        Option 3 = 'b': `lov` -> B-TAG, `###ing` -> B-TAG
        """
        assert tag_for_subtokens in [
            None,
            "i",
            "b",
        ], "Param :tag_for_subtokens must be one of `None`, `'i'`, `'b'`!"
        new_labels = []
        current_word = None
        for word_id in word_ids:
            if word_id is None:
                # Special token
                new_labels.append(-100)
            elif word_id != current_word:
                # Start of a new word!
                current_word = word_id
                new_labels.append(labels[word_id])
            else:
                # Same word as previous token
                if tag_for_subtokens is not None:
                    label = labels[word_id]
                    # If the label is B-XXX we change it to I-XXX
                    if tag_for_subtokens == "i" and label % 2 == 1:
                        label += 1
                else:
                    label = -100
                new_labels.append(label)

        return new_labels

    """
    Returns a function that converts text to list of ids (of tokens)
    """

    @staticmethod
    def get_tokenize_function(
        tokenizer,
        text_name,
        label_name,
        test_mode: bool = False,
        save_first_bpe_mask: bool = True,
        **kwargs
    ):
        if test_mode:

            def tokenize_function(examples):
                tokens = examples[text_name]
                tokenized_inputs = tokenizer(
                    tokens, truncation=True, is_split_into_words=True
                )
                # Save ids of the first BPEs if necessary
                if save_first_bpe_mask:
                    idx_first_bpe = []
                    for i in range(len(tokens)):
                        word_ids = tokenized_inputs.word_ids(i)
                        idx_first_bpe.append(
                            [
                                i
                                for (i, x) in enumerate(word_ids[1:], 1)
                                if ((x != word_ids[i - 1]) and (x is not None))
                            ]
                        )
                    tokenized_inputs["idx_first_bpe"] = idx_first_bpe
                return tokenized_inputs

        else:

            def tokenize_function(examples):
                tokenized_inputs = tokenizer(
                    examples[text_name], truncation=True, is_split_into_words=True
                )
                if not test_mode:
                    all_labels = examples[label_name]
                    new_labels = []
                    idx_first_bpe = []
                    for i, labels in enumerate(all_labels):
                        word_ids = tokenized_inputs.word_ids(i)
                        new_labels.append(
                            WrapperNer.align_labels_with_tokens(labels, word_ids)
                        )
                        # Save ids of the first BPEs if necessary
                        if save_first_bpe_mask:
                            idx_first_bpe.append(
                                [
                                    i
                                    for (i, x) in enumerate(word_ids[1:], 1)
                                    if ((x != word_ids[i - 1]) and (x is not None))
                                ]
                            )
                    tokenized_inputs["labels"] = new_labels
                    if save_first_bpe_mask:
                        tokenized_inputs["idx_first_bpe"] = idx_first_bpe
                return tokenized_inputs

        return tokenize_function

    @staticmethod
    def get_model_class():
        return AutoModelForTokenClassification

    @staticmethod
    def get_additional_tokenization_train_kwargs():
        return {"save_first_bpe_mask": False}

    @staticmethod
    def get_additional_tokenization_inference_kwargs():
        return {"save_first_bpe_mask": True}

    @staticmethod
    def get_data_collator_class():
        return DataCollatorForTokenClassification

    @staticmethod
    def get_datasets_metric_name():
        return "seqeval"

    @staticmethod
    def get_metric_name():
        return "test_overall_f1"

    def get_kwargs_for_compute_metrics_fn(self):
        return {
            "id2label": self.id2label,
            "display_tag_stats": "f1",
            "padding_idx": -100,
        }

    @staticmethod
    def get_compute_metrics_function(
        metric,
        id2label: Union[Dict[int, str], List],
        additional_metrics: Union[Tuple[Metric], List[Metric]] = tuple(),
        display_tag_stats="f1",
        padding_idx=-100,
    ):
        def compute_metrics(eval_preds):
            logits_ids, labels_ids, *inputs = eval_preds
            preds_ids = logits_ids.argmax(axis=-1)
            # Convert ids to tags
            idx_first_bpe = np.array(labels_ids) != padding_idx
            # Remove padding
            labels_without_padding = WrapperNer._remove_padding(
                labels_ids, idx_first_bpe
            )
            preds_without_padding = WrapperNer._remove_padding(preds_ids, idx_first_bpe)
            # Convert idx to tokens
            labels = WrapperNer._convert_id2label(labels_without_padding, id2label)
            preds = WrapperNer._convert_id2label(preds_without_padding, id2label)
            # Compute metrics
            metrics_dict = metric.compute(predictions=preds, references=labels)
            for add_metric in additional_metrics:
                add_metric_dict = add_metric.compute(
                    predictions=preds, references=labels
                )
                metrics_dict.update(add_metric_dict)
            # Individual entity scores are presented as dicts now. Need to "straighten" them
            output_dict = WrapperNer._construct_metrics_dict_for_ner(
                metrics_dict, display_tag_stats
            )
            return output_dict

        return compute_metrics

    @staticmethod
    def _convert_id2label(labels_or_preds, id2label, padding_idx=-100):
        return [
            [id2label[l] for l in label if l != padding_idx]
            for label in labels_or_preds
        ]

    @staticmethod
    def _construct_metrics_dict_for_ner(
        metrics_dict, display_tag_stats: str or bool = "f1"
    ):
        output_dict = {}
        for key, val in metrics_dict.items():
            if isinstance(val, dict):
                if display_tag_stats is False or display_tag_stats is None:
                    continue
                elif display_tag_stats == "f1":
                    output_dict[key + "_f1"] = val["f1"]
                else:
                    for key_2, val_2 in val.items():
                        output_dict[key + "_" + key_2] = val_2
            else:
                output_dict[key] = val

        return output_dict

    @staticmethod
    def _remove_padding(predictions, idx_first_bpe):
        preds = []
        for pred, cond in zip(predictions, idx_first_bpe):
            if isinstance(pred, Tensor):
                pred = pred.cpu().detach().numpy()
            pred_without_padding = pred[cond]
            preds.append(pred_without_padding)

        return np.array(preds)



import logging
from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import calculate_mahalanobis_distance
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def mahalanobis_sampling(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **mahalanobis_kwargs,
):
    batched = mahalanobis_kwargs.get("mahalanobis_batched", False)
    substrategy = mahalanobis_kwargs.get("mahalanobis_substrategy", "lc")
    use_class_probas = mahalanobis_kwargs.get("mahalanobis_use_class_probas", False)
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        train_data=X_train,
        unlabeled_data=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        # Mahalanobis
        classwise=True,  # use an own centroid for each class
        batched=batched,
        use_v2=mahalanobis_kwargs.get("mahalanobis_use_v2", False),
        use_activation=mahalanobis_kwargs.get("mahalanobis_use_activation", False),
        use_spectralnorm=mahalanobis_kwargs.get("mahalanobis_use_spectralnorm", True),
    )
    if not batched:
        dists = calculate_mahalanobis_distance(**kwargs)
        if use_class_probas:
            log.info("Using probas for mahalanobis")
            label_name = model.data_config["label_name"]
            class_probas = np.bincount(
                X_train[label_name], minlength=model.num_labels
            ) / len(X_train)
            dists = dists * class_probas
        else:
            log.info("Not using probas for mahalanobis))")
        if substrategy == "lc":
            uncertainty_estimates = np.min(dists, axis=-1)
        elif substrategy == "margin":
            dists.sort(axis=1)
            max_dists = dists[:, 0]
            second_max_dists = dists[:, 1]
            uncertainty_estimates = (
                max_dists - second_max_dists
            )  # the greater the value, the more uncertain we are
        else:
            raise NotImplementedError
        argsort = np.argpartition(-uncertainty_estimates, kth=n_instances)
        query_idx = argsort[:n_instances]
    else:
        query_idx, uncertainty_estimates = calculate_mahalanobis_distance(
            n_instances=n_instances, **kwargs
        )
    query = X_pool.select(query_idx)
    return query_idx, query, uncertainty_estimates



import numpy as np


def get_ups_sampling_probas(argsort, gamma, T):
    ranks = argsort.argsort() / len(argsort)
    return np.exp(-np.maximum(0, ranks - gamma) / np.maximum(T, 1e-8))


def sample_idxs(sampling_probas):
    to_select = []
    for i in range(len(sampling_probas)):
        proba_to_choose_i = sampling_probas[i]
        if np.random.uniform() < proba_to_choose_i:
            to_select.append(i)
    return np.array(to_select)


def ups_subsampling(uncertainty_estimates, gamma_or_k_confident_to_save, T):
    if isinstance(gamma_or_k_confident_to_save, int):
        gamma_or_k_confident_to_save /= len(uncertainty_estimates)
    argsort = np.argsort(-uncertainty_estimates)
    sampling_probas = get_ups_sampling_probas(argsort, gamma_or_k_confident_to_save, T)
    return sample_idxs(sampling_probas)



import logging
from pathlib import Path
from time import time

from .query_strategies.al_strategy_utils import take_idx, concatenate_data
from .utils.general import json_dump

logger = logging.getLogger()


class ActiveLearner:
    def __init__(
        self,
        estimator,
        query_strategy,
        train_data,  # labeled_data
        strategy_kwargs=None,
        sampling_strategy=None,
        sampling_kwargs=None,
        log_dir="logs/",
        **kwargs
    ):
        self.estimator = estimator
        self.query_strategy = query_strategy
        self.train_data = train_data
        self.strategy_kwargs = strategy_kwargs if strategy_kwargs is not None else {}
        self.sampling_strategy = sampling_strategy
        self.sampling_kwargs = sampling_kwargs if sampling_kwargs is not None else {}

        self._create_time_dict(log_dir)

        self._fit_estimator()  # TODO: remove and use start instead

    def _fit_estimator(self, data=None):
        start_time = time()
        if data is not None:
            self.estimator.fit(data)
        else:
            self.estimator.fit(self.train_data)
        fit_time = time() - start_time
        self._add_obs_to_time_dict("fit", fit_time)

    def _add_data(self, new_data):
        self.train_data = concatenate_data(self.train_data, new_data)

    def query(self, X_pool, n_instances, **kwargs):
        strategy_kwargs = dict(self.strategy_kwargs)
        strategy_kwargs["X_train"] = self.train_data
        strategy_kwargs.update(kwargs)

        start_time = time()
        output = self.query_strategy(
            self.estimator, X_pool, n_instances, **strategy_kwargs
        )
        query_time = time() - start_time
        self._add_obs_to_time_dict("query", query_time)

        return output

    def teach(self, data, fit_only_new=False):
        self._add_data(data)
        if fit_only_new:
            self._fit_estimator(data)
        else:
            self._fit_estimator()

    def sample(self, X_pool, uncertainty_estimates, gamma_or_k_confident_to_save=None):
        sampling_kwargs = dict(self.sampling_kwargs)
        if gamma_or_k_confident_to_save is not None:
            sampling_kwargs[
                "gamma_or_k_confident_to_save"
            ] = gamma_or_k_confident_to_save

        sample_idx = self.sampling_strategy(uncertainty_estimates, **sampling_kwargs)
        sample_instances = take_idx(X_pool, sample_idx)
        return sample_idx, sample_instances

    def _create_time_dict(self, log_dir: str or Path):
        self._time_dict = {"fit": [], "query": []}
        self._time_dict_path = Path(log_dir) / "al_learner_time_dict.json"
        if not Path(log_dir).exists():
            Path(log_dir).mkdir()
        json_dump(self._time_dict, self._time_dict_path)

    def _add_obs_to_time_dict(self, step: str, value: float):
        self._time_dict[step].append(value)
        json_dump(self._time_dict, self._time_dict_path)



import os

from setuptools import setup, find_packages

requirements = [
    "hydra-core==1.1.0",
    "datasets>=2.3.2",
    "rouge-score==0.0.4",
    "plotly==5.5.0",
    "psutil",
    "kaleido==0.2.1",
    "seqeval==1.2.2",
    "nlpaug>=1.1.10",
    "ruamel.yaml==0.17.20",
    "scikit-learn==1.0.2",
    "protobuf==3.19.4",
    "tqdm",
    "matplotlib",
    "pandas",
    "mlflow==1.7.2",
    "KDEpy==1.1.0",
    "hnswlib==0.6.0",
    "aioredis",
    "torch",
    "bs4",
    "ray==1.9.1",
    "transformers>=4.20.1",
    "nltk==3.6.5",
    "sacrebleu==1.5.0",
    "pytest==7.1.2",
    "pytest-cov==3.0.0",
    "toma==1.1.0",
    "pytest-runner",
    "hf-lfs>=0.0.3",
    "fastcluster",
    "thinc==8.0.12",
    "wget",
    "gensim",
    "pytreebank",
    "flair==0.10",
    "yargy",
    "yaspin",
    "rich",
    "spacy",
    "regex",
    "pybind11==2.10.0",
    "ipykernel==6.15.1",
    "small-text",
    "genbadge",
    "tensorboardX",
    "tensorboard",
]

os.system("chmod a+x init.sh examples/*")

setup(
    name="acleto",
    packages=find_packages(include=["acleto*"]),
    version="0.0.7",
    description="A Library for active learning. Supports text classification and sequence tagging tasks.",
    author="Tsvigun A., Sanochkin L., Kuzmin G., Larionov D., and Dr Shelmanov A.",
    license="MIT",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    keywords="NLP active AL deep learning transformer pytorch PLASM UPS",
    install_requires=requirements,
    extras_require={"small-text": ["small-text"], "modAL": ["modAL"],},
    include_package_data=True,
    setup_requires=["pytest-runner"],
    tests_require=["pytest==7.1.2"],
    test_suite="tests",
)

os.system("./init.sh")



import logging
from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset
from scipy import stats
from sklearn.metrics import pairwise_distances

import torch
from torch import nn

from ..utils.get_embeddings import get_grad_embeddings
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def badge(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    **badge_kwargs,
):
    """
    Measures uncertainty as the gradient magnitude with respect to parameters in the final (output) layer,
    which is computed using the most likely label according to the model. To capture diversity, collect a
    batch of examples where these gradients span a diverse set of directions. https://arxiv.org/abs/1906.03671
    """
    logits = model.predict_logits(X_pool)

    kwargs = dict(
        # Necessary
        model_wrapper=model,
        data_test=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=False,
        logits=logits,
    )

    vectors = calculate_badge_scores(**kwargs).cpu().detach().numpy()

    query_idx = np.array(init_centers(vectors, k=n_instances))

    query = X_pool.select(query_idx)

    # Uncertainty estimates are not defined for BADGE
    uncertainty_estimates = np.zeros(len(X_pool))

    return query_idx, query, uncertainty_estimates


def init_centers(x, k):
    ind = np.argmax([np.linalg.norm(s, 2) for s in x])
    mu = [x[ind]]
    inds_all = [ind]
    cent_inds = [0.0] * len(x)
    cent = 0
    while len(mu) < k:
        if len(mu) == 1:
            d2 = pairwise_distances(x, mu).ravel().astype(float)
        else:
            new_d = pairwise_distances(x, [mu[-1]]).ravel().astype(float)
            for i in range(len(x)):
                if d2[i] > new_d[i]:
                    cent_inds[i] = cent
                    d2[i] = new_d[i]
        d2 = d2.ravel().astype(float)
        d_dist = (d2 ** 2) / sum(d2 ** 2)
        custom_dist = stats.rv_discrete(
            name="custm", values=(np.arange(len(d2)), d_dist)
        )
        ind = custom_dist.rvs(size=1)[0]
        while ind in inds_all:
            ind = custom_dist.rvs(size=1)[0]
        mu.append(x[ind])
        inds_all.append(ind)
        cent += 1
    return inds_all

def calculate_badge_scores(
    model_wrapper,
    data_test,
    logits,
    use_activation: bool = False,
    use_spectralnorm: bool = False,
    data_is_tokenized=False,
    data_config=None,
    batch_size=100,
    to_numpy=True,
):
    data_config = data_config if data_config is not None else model_wrapper.data_config
    kwargs = dict(
        # General
        model=model_wrapper.model,
        prepare_model=True,
        batch_size=batch_size,
        to_numpy=False,
        data_is_tokenized=data_is_tokenized,
        tokenizer=model_wrapper.tokenizer,
        task=model_wrapper.task,
        text_name=data_config["text_name"],
        label_name=data_config["label_name"],
    )
    """Return the loss gradient with respect to the penultimate layer for BADGE"""
    pooled_output = get_grad_embeddings(dataloader_or_data=data_test, **kwargs).to(
        model_wrapper.model.device
    )
    batch_size, num_classes = logits.shape
    probs = nn.functional.softmax(
        torch.Tensor(logits).to(model_wrapper.model.device), dim=-1
    )
    preds = probs.argmax(dim=1)
    preds_oh = nn.functional.one_hot(preds, num_classes=num_classes)
    preds_oh = preds_oh.type(torch.cuda.FloatTensor)
    scales = probs - preds_oh
    grads_3d = torch.einsum("bi,bj->bij", scales, pooled_output)
    grads = grads_3d.view(batch_size, -1)
    return grads



import numpy as np
import torch
from torch.utils.data import DataLoader
from tqdm.notebook import tqdm

from ..model_wrappers.transformers import WrapperCls, WrapperNer, WrapperAts, WrapperNmt

WRAPPERS = {"cls": WrapperCls, "ner": WrapperNer, "ats": WrapperAts, "nmt": WrapperNmt}


def compute_egl_scores(
    model,
    dataloader_or_data,
    data_is_tokenized=False,
    batch_size=100,
    **tokenization_kwargs,
):
    device = next(model.model.parameters()).device
    criterion = torch.nn.CrossEntropyLoss(reduction="none").to(device)

    wrapper = WRAPPERS[tokenization_kwargs.pop("task", "cls")]
    if not isinstance(dataloader_or_data, DataLoader):
        if not data_is_tokenized:
            if "label_name" not in tokenization_kwargs:
                tokenization_kwargs["test_mode"] = True
            tokenize_function = wrapper.get_tokenize_function(
                data=dataloader_or_data,
                **tokenization_kwargs,  # WTF: This should not be binded to the particular Wrapper
            )
            columns_to_remove = [
                x
                for x in dataloader_or_data.features.keys()
                if x not in ["labels", "weight"]
            ]
            # The last two arguments is a temporary fix for Kristofari
            dataloader_or_data = dataloader_or_data.map(
                tokenize_function,
                batched=True,
                remove_columns=columns_to_remove,
                load_from_cache_file=False,
                cache_file_name=f"tmp.data",
            )
            if wrapper == WrapperNer:
                idx_first_bpe = dataloader_or_data["idx_first_bpe"]
                dataloader_or_data = dataloader_or_data.remove_columns("idx_first_bpe")
        dataloader_or_data = DataLoader(
            dataloader_or_data,
            shuffle=False,
            batch_size=batch_size,
            collate_fn=wrapper.get_data_collator_class()(
                tokenizer=tokenization_kwargs["tokenizer"]
            ),
            pin_memory=(str(device).startswith("cuda")),
        )

    num_obs = len(dataloader_or_data.dataset)

    gradient_lengths = initialize_gradient_lengths_array(num_obs)

    offset = 0
    possible_input_keys = model.model.forward.__code__.co_varnames
    possible_input_keys = list(possible_input_keys) + ["input_ids", "attention_mask"]

    for batch in tqdm(dataloader_or_data):
        batch = {k: v.to(device) for k, v in batch.items() if k in possible_input_keys}
        compute_gradient_lengths(
            model, criterion, gradient_lengths, offset, batch, device
        )

        offset += len(batch["input_ids"])

    return gradient_lengths


def initialize_gradient_lengths_array(dim):
    return np.zeros(dim, dtype=np.float64)


def compute_gradient_lengths(model, criterion, gradient_lengths, offset, x, device):
    batch_len = len(x["input_ids"])
    all_classes = torch.LongTensor([batch_len * [i] for i in range(model.num_labels)])
    if device is not None and device != "cpu":
        all_classes = all_classes.to(device)

    gradients = initialize_gradients(batch_len, model.num_labels, device)

    # x = x.to(device, non_blocking=True)
    model.model.zero_grad()

    compute_gradient_lengths_batch(model, criterion, x, gradients, all_classes)
    aggregate_gradient_lengths_batch(batch_len, gradient_lengths, gradients, offset)


def initialize_gradients(batch_len, num_classes, device):
    return torch.zeros([num_classes, batch_len]).to(device, non_blocking=True)


def compute_gradient_lengths_batch(model, criterion, x, gradients, all_classes):

    batch_len = len(x["input_ids"])

    output = model.model(**x)["logits"]
    with torch.no_grad():
        sm = torch.nn.functional.softmax(output, dim=1)

    for j in range(model.num_labels):
        loss = criterion(output, all_classes[j])

        for k in range(batch_len):
            model.model.zero_grad()
            loss[k].backward(retain_graph=True)

            compute_gradient_length(model, sm, gradients, j, k)


def compute_gradient_length(model, sm, gradients, j, k):

    params = [
        param.grad.flatten()
        for param in model.model.parameters()
        if param.requires_grad
    ]
    params = torch.cat(params)

    gradients[j, k] = torch.sqrt(params.pow(2)).sum()
    gradients[j, k] *= sm[k, j].item()


def aggregate_gradient_lengths_batch(batch_len, gradient_lengths, gradients, offset):
    gradient_lengths[offset : offset + batch_len] = torch.sum(gradients, 0).to(
        "cpu", non_blocking=True
    )



import json
import logging
from pathlib import Path
from numpy import nan

import pandas as pd
import yaml

log = logging.getLogger()

NUM_QUERIES = 16
QUANTILES = (-1, -1.0, 0.01, 0.025, 0.05, 0.1)
METRIC_NAME = "accuracy"  # f1-measure-overall / "rougeL" / "BLEU"

METRIC_FILE_NAMES = [
    "target_metrics.json",
    "successor_metrics.json",
    "acquisition_metrics.json",
    "metrics.json",
]

DEFAULT_METRIC = {
    "cls": "accuracy",
    "ner": "overall_f1",
    "ats": "rougeL",
    "nmt": "bleu",
}

AL_STRATEGIES = ["random", "mahalanobis", "nuq", "logits", "ddu", "margin", "oracle"]


def _json_load_with_unknown_prefix(file_name, prefixes, metric_name):
    with open(file_name) as f:
        metrics = json.load(f)
    for prefix in prefixes:
        try:
            target_metric = metrics[prefix + metric_name]
            return target_metric
        except:
            continue


def _add_observation_to_df(
    df, exp_data, metrics, num_queries, id_experiment, fill_missing=False
):
    row = (
        [id_experiment]
        + [exp_data[x] for x in df.columns[1:10]]
        + metrics[:num_queries]
    )
    if len(row) < df.shape[1] and fill_missing:
        row = row + [None for _ in range(df.shape[1] - len(row))]
    df.loc[len(df)] = row


def extract_info_from_config(config):
    data = {}
    data["strategy"] = config["al"]["strategy"] if "al" in config else "full"
    data["acquisition"] = config["acquisition_model"].get("checkpoint") or config[
        "acquisition_model"
    ].get("name")
    data["successor"] = (
        config["successor_model"].get("checkpoint")
        or config["successor_model"].get("name")
        if config["successor_model"]
        else data["acquisition"]
    )
    data["target"] = (
        config["target_model"].get("checkpoint") or config["target_model"].get("name")
        if "target_model" in config
        else data["successor"]
    )
    data["deleted"] = (
        config["al"]["share_confident_to_save"]
        if "al" in config and "share_confident_to_save" in config["al"]
        else round(1 - config["al"]["plasm_thresh"], 4)
        if "al" in config
        and config["al"].get("plasm_thresh", 1) != 1
        and isinstance(config["al"]["plasm_thresh"], (int, float))
        else config["al"]["plasm_thresh"]
        if "al" in config and isinstance(config["al"].get("plasm_thresh", 1), str)
        else 0
    )
    data["seed"] = config["seed"]
    dataset = config["data"]["dataset_name"]
    data["dataset"] = " ".join(dataset) if isinstance(dataset, list) else dataset
    data["ups"] = (
        "None"
        if "al" not in config
        else "k=1.0, T=0"
        if not config["al"]["sampling_type"]
        else f"k={config['al']['gamma_or_k_confident_to_save']}, T={config['al']['T']}"
    )
    return data


def search_for_experiments(
    path,
    df,
    num_queries=16,
    metric_name="accuracy",
    fill_missing=False,
    prefix: str or None = "test",  # "test" or "eval" or None
    quantiles=QUANTILES,
    log_errors=False,
    metric_file_names=None,
    id_experiment=0,
    exclude_cache_from_dir: bool = True
):
    if exclude_cache_from_dir and "cache" in str(path):
        return
    if prefix is None or prefix == "none":
        metrics_prefix = ""
    else:
        metrics_prefix = prefix + "_"

    if metric_file_names is None:
        metric_file_names = METRIC_FILE_NAMES
    elif isinstance(metric_file_names, str):
        metric_file_names = [metrics_prefix]

    iterdir = list(path.iterdir())
    if (path / "config.yaml" in iterdir) and (
        (path / "acquisition_metrics.json" in iterdir)
        or (path / "metrics.json" in iterdir)
        or any((path / file in iterdir for file in metric_file_names))
    ):

        try:
            with open(path / "config.yaml") as f:
                config = yaml.load(f, yaml.Loader)

            for file_name in metric_file_names:
                if path / file_name in iterdir:
                    metrics = _json_load_with_unknown_prefix(
                        path / file_name,
                        [metrics_prefix, "test_", "eval_", ""],
                        metric_name,
                    )
                    break

            if "tracin" in config and config["tracin"]["use"]:
                tracin_results = []
                processed_quantiles = []
                for quant in quantiles:
                    try:
                        tracin_results.append(
                            _json_load_with_unknown_prefix(
                                path / f"target_tracin_quantile_{quant}_metrics.json",
                                [metrics_prefix, "test_", "eval_", ""],
                                metric_name,
                            )
                        )
                        processed_quantiles.append(quant)
                    except Exception as e:
                        if log_errors:
                            log.error(e, exc_info=True)
                quantiles = processed_quantiles

            al_type = "one"  # Default value
            if "acquisition_model" not in config:
                config["acquisition_model"] = config["model"]
                config["successor_model"] = None
                metrics = [metrics for _ in range(num_queries)]
                al_type = "full"
            else:
                if "target_model" in config:
                    if config["al"].get("plasm_thresh", 1) != 1:
                        al_type = "plasm_thresh"
                    else:
                        al_type = "plasm"
                elif config["successor_model"]:
                    al_type = "asm"

            exp_data = extract_info_from_config(config)
            exp_data["al_type"] = al_type
            _add_observation_to_df(
                df, exp_data, metrics, num_queries, id_experiment, fill_missing
            )

            if "tracin" in config and config["tracin"]["use"]:
                for i, metrics in enumerate(tracin_results):
                    exp_data["deleted"] = quantiles[i]
                    _add_observation_to_df(
                        df, exp_data, metrics, num_queries, id_experiment, fill_missing
                    )

        except Exception as e:
            if log_errors:
                if (
                    isinstance(e, ValueError)
                    and str(e) == "cannot set a row with mismatched columns"
                ):
                    log.error(
                        f"Path: {str(path)}; num obs: {len(metrics[:num_queries])}; "
                        f"required num obs: {df.shape[1] - len([exp_data[x] for x in df.columns[:10]])}"
                    )
                else:
                    log.error(e, exc_info=True)
            return

    else:
        for file in path.iterdir():
            if file.is_dir():
                search_for_experiments(
                    file,
                    df,
                    num_queries,
                    metric_name,
                    fill_missing=fill_missing,
                    prefix=prefix,
                    quantiles=quantiles,
                    log_errors=log_errors,
                    metric_file_names=metric_file_names,
                    id_experiment=id_experiment,
                )


def collect_data(
    paths,
    num_queries=16,
    task="cls",
    fill_missing=False,
    prefix: str or None = "test",  # "test" or "eval" or None
    quantiles=QUANTILES,
    log_errors=False,
    metric_name=None,
    metric_file_names=None,
    exclude_cache_from_dir: bool = True
):
    if metric_name is None:
        metric_name = DEFAULT_METRIC[task]

    if quantiles is None:
        quantiles = QUANTILES

    df = pd.DataFrame(
        columns=[
            "id_experiment",
            "al_type",
            "strategy",
            "dataset",
            "acquisition",
            "successor",
            "target",
            "deleted",
            "ups",
            "seed",
        ]
        + [f"iter_{i}" for i in range(num_queries)]
    )

    [
        search_for_experiments(
            Path(path),
            df=df,
            num_queries=num_queries,
            metric_name=metric_name,
            fill_missing=fill_missing,
            prefix=prefix,
            quantiles=quantiles,
            log_errors=log_errors,
            metric_file_names=metric_file_names,
            id_experiment=id_experiment,
            exclude_cache_from_dir=exclude_cache_from_dir
        )
        for id_experiment, path in enumerate(paths)
    ]
    df.loc[:, "iter_0":] = df.loc[:, "iter_0":].astype(float)
    df = df[~df.duplicated()]
    # Replace values for abs-sum metrics
    if (
        metric_name.startswith("ngram_overlap")
        or metric_name.startswith("novel_ngrams_2")
        or metric_name.startswith("cons_score")
    ):
        df.loc[:, "iter_0":] = df.loc[:, "iter_0":].replace(-1, nan)
    log.info(f"Num observations: {len(df)}")
    return df



import logging

import numpy as np
import pandas as pd

logger = logging.getLogger()


class ActiveLearner:
    """A class that implements the active learning logic."""

    def __init__(
        self,
        active_learn_alg,
        X_labeled_dataset,
        y_labeled_dataset,
        X_unlabeled_dataset,
        y_unlabeled_dataset=None,
        model_evaluate=None,
        X_test_dataset=None,
        y_test_dataset=None,
        eval_metrics=None,
        rnd_start_steps=0,
        n_instances=10,
    ):
        """
        ActiveLearner constructor.

        Args:
            active_learn_alg (functor): query strategy wrapper.
            X_full_dataset (np.array or sparse matrix): feature matrix.
            y_dtype: type of y labels.
            y_full_dataset (np.array): known answers (e.g., None -- unknown, True -- positive class, False -- negative class)
            model_evaluate: the model that will be evaluated on the holdout.
            X_test_dataset: feature matrix for testing via holdout.
            y_test_dataset: y labels for testing via holdout.
            eval_metrics (list): list of sklearn evaluation metrics.
            rnd_start_steps: AL will can make several seed steps by choosing random samples (without model suggestions).
            logger (logging.Logger): the object for logging.

        """
        super().__init__()

        self._X_unlabeled_dataset = X_unlabeled_dataset
        self._X_labeled_dataset = X_labeled_dataset
        self._y_labeled_dataset = y_labeled_dataset
        if y_unlabeled_dataset is not None:
            self._y_unlabeled_dataset = y_unlabeled_dataset
        else:
            self._y_unlabeled_dataset = [None] * len(self._X_unlabeled_dataset)

        self._active_learn_alg = active_learn_alg

        self._X_test_dataset = X_test_dataset
        self._y_test_dataset = y_test_dataset
        self._model_evaluate = model_evaluate
        self._eval_metrics = eval_metrics

        self._iteration_num = 0
        self._rnd_start_steps = rnd_start_steps

        self.n_instances = n_instances

    def _select_unannotated(self, labels):
        return np.where([(e is None) for e in labels])[0]

    def start(self):
        self._active_learn_alg.start(
            self._X_labeled_dataset,
            self._y_labeled_dataset,
            self._X_unlabeled_dataset,
            self._y_unlabeled_dataset,
        )

    def choose_random_sample_for_annotation(self, number):
        #         if len(np.where([(e is None) for e in self._y_unlabeled_dataset])[0]) == len(
        #             self._X_unlabeled_dataset
        #         ):

        #         else:
        #             ids_array = np.arange(len(self._X_unlabeled_dataset))
        # TODO: fix
        ids_array = self._select_unannotated(self._y_unlabeled_dataset)
        return np.random.choice(ids_array, size=number, replace=False)

    def choose_samples_for_annotation(self):
        if self._iteration_num < self._rnd_start_steps:
            return self.choose_random_sample_for_annotation(self.n_instances)
        else:
            return self._active_learn_alg.choose_samples_for_annotation(
                self.n_instances
            )

    def evaluate(self, fit_model=True):
        if self._model_evaluate is None:
            return None

        if fit_model:
            selector = [n for n, e in enumerate(self._y_full_dataset) if e is not None]
            y_fit = [self._y_full_dataset[i] for i in selector]
            X_fit = [self._X_full_dataset[i] for i in selector]
            logger.info("Number of training samples: {}".format(len(y_fit)))

            self._model_evaluate.fit(X_fit, y_fit)

        preds = self._model_evaluate.predict(self._X_test_dataset)
        return {
            metric.__name__: metric(preds, self._y_test_dataset)
            for metric in self._eval_metrics
        }

    def get_annotation(self):
        return self._y_unlabeled_dataset

    def make_iteration(self, indexes, answers):
        self.add_answers(indexes, answers)
        res = self._active_learn_alg.make_iteration(indexes, answers)
        self._iteration_num += 1
        return res

    def add_answers(self, indexes, answers):
        answers = pd.Series(answers, index=indexes, dtype="object")
        answers = answers[answers.notnull()]
        logger.info(f"Iteration Answers: {answers.to_dict()}")
        for num, i in enumerate(answers.index):
            self._y_unlabeled_dataset[i] = answers[i]

    def is_busy(self):
        return False



import logging
import os
import shutil
from copy import deepcopy

import numpy as np
import torch
import wget
from gensim.models.keyedvectors import KeyedVectors
from tokenizers.pre_tokenizers import Whitespace

log = logging.getLogger()


def load_word_vectors(
    vectors_name, embeddings_path, model_cache_dir=None, n_vectors=None
):
    """Loads embeddings by url and name."""
    if vectors_name == "fasttext":
        embeddings_name = embeddings_path.split("/")[-1]
        save_path = os.path.join(model_cache_dir, embeddings_name)
        # check if this file already loaded
        os.makedirs(model_cache_dir, exist_ok=True)
        if not (os.path.isfile(save_path)):
            save_path = wget.download(embeddings_path, out=save_path)
        # unzip it and extract data to arrays
        if not (os.path.isfile(os.path.join(model_cache_dir, "crawl-300d-2M.vec"))):
            shutil.unpack_archive(save_path, model_cache_dir)
        fname = "crawl-300d-2M.vec"
    elif vectors_name == "word2vec":
        embeddings_name = "GoogleNews-vectors-negative300.bin.gz"
        save_path = os.path.join(model_cache_dir, embeddings_name)
        # check if this file already loaded
        os.makedirs(model_cache_dir, exist_ok=True)
        if not (os.path.isfile(save_path)):
            save_path = wget.download(embeddings_path, out=save_path)
        if not os.path.isfile(
            os.path.join(model_cache_dir, "GoogleNews-vectors-negative300.txt")
        ):
            model = KeyedVectors.load_word2vec_format(save_path, binary=True)
            model.save_word2vec_format(
                os.path.join(model_cache_dir, "GoogleNews-vectors-negative300.txt"),
                binary=False,
            )
        fname = "GoogleNews-vectors-negative300.txt"
    embeddings = []
    word2idx = {}
    log.info(f"Loading embeddings...")
    with open(os.path.join(model_cache_dir, fname), "r") as f:
        for idx, line in enumerate(f.readlines()):
            if idx == 0:
                # 0 line contains number of words
                continue
            tokens = line.rstrip().split(" ")
            word = tokens[0]
            vector = np.array(tokens[1:], dtype=np.float32)
            # here each word is unique, because we load pretrained embeddings
            word2idx[word] = idx
            embeddings.append(vector)
            if n_vectors is not None and idx > n_vectors:
                # use n_vectors parameter to load first n_vectors
                # mostly used for tests and debug
                break
    # add pad and unk tokens
    word2idx["[UNK]"] = 0
    embeddings.insert(0, np.mean(embeddings, axis=0))
    word2idx["[PAD]"] = idx + 1
    embeddings.append(np.zeros_like(embeddings[-1]))
    embeddings = torch.Tensor(np.asarray(embeddings))
    # After map words to ids and add pad and unk tokens
    # and also return np array of embeddings, not dict
    return embeddings, word2idx


def load_embeddings(embeddings_path, model_cache_dir=None, n_vectors=None):
    if embeddings_path == "fasttext":
        return load_word_vectors(
            "fasttext",
            "https://dl.fbaipublicfiles.com/fasttext/vectors-english/crawl-300d-2M.vec.zip",
            model_cache_dir,
            n_vectors,
        )
    elif embeddings_path == "word2vec":
        return load_word_vectors(
            "word2vec",
            "https://drive.google.com/uc?id=0B7XkCwpI5KDYNlNUTTlSS21pQmM&confirm=no_antivirus",
            model_cache_dir,
            n_vectors,
        )
    else:
        raise NotImplementedError


def check_models(config):
    """Check if any of a models is a cnn or a custom model, and get params of the model"""
    from ..models import PYTORCH_INIT_MODELS_DICT

    models_type = ["model", "acquisition_model", "successor_model", "target_model"]
    for model in models_type:
        if config.get(model, False):
            if config[model]["checkpoint"] in PYTORCH_INIT_MODELS_DICT.keys() or config[
                model
            ]["checkpoint"].endswith(".py"):
                return (
                    config[model]["embeddings_path"],
                    config[model]["embeddings_cache_dir"],
                )
    return None, None


def load_embeddings_if_necessary(*datasets, config):
    from datasets import concatenate_datasets

    embeddings, word2idx = None, None
    embeddings_path, embeddings_cache_dir = check_models(config)
    if embeddings_path is not None:
        # load embeddings
        try:
            all_data = concatenate_datasets([*datasets])
        except:
            all_data = deepcopy(datasets[0])
            for i in range(1, len(datasets)):
                all_data.add(datasets[i])
        if "model" in config.keys():
            embeddings_path = config.model.embeddings_path
            embeddings_cache_dir = config.model.embeddings_cache_dir
        else:
            embeddings_path = config.acquisition_model.embeddings_path
            embeddings_cache_dir = config.acquisition_model.embeddings_cache_dir
        embeddings, word2idx = load_embeddings_with_text(
            all_data,
            embeddings_path,
            embeddings_cache_dir,
            text_name=config.data.text_name,
            n_vectors=config.data.get("n_vector", None),
        )
    return embeddings, word2idx


def load_embeddings_with_text(
    data,
    embeddings_path="https://dl.fbaipublicfiles.com/fasttext/vectors-english/crawl-300d-2M.vec.zip",
    model_cache_dir=None,
    text_name="text",
    n_vectors=None,
):
    """Loads embeddings by url, uses fasttext by default.
    Now works only for fasttext. Also only use words from data for embeddings
    """
    embeddings, word2idx = load_embeddings(embeddings_path, model_cache_dir, n_vectors)
    # now reduce embeddings size
    pre_tokenizer = Whitespace()
    new_word2idx = {}
    new_embeddings = []
    new_word2idx["[UNK]"] = 0
    idx = 1
    for sentence in data:
        # split sentence by words with same pretokenizer
        if type(sentence[text_name]) == str:
            tokens = pre_tokenizer.pre_tokenize_str(sentence[text_name])
        else:
            tokens = sentence[text_name]
        tokens = [token[0].lower() for token in tokens]
        for token in tokens:
            if token not in new_word2idx and token in word2idx:
                new_word2idx[token] = idx
                new_embeddings.append(
                    np.array(embeddings[word2idx[token]], dtype=np.float32)
                )
                idx += 1
            elif token not in new_word2idx and token not in word2idx:
                # use random vector for embedding
                embedding = np.random.uniform(-0.1, 0.1, embeddings[0].shape)
                new_word2idx[token] = idx
                new_embeddings.append(np.array(embedding, dtype=np.float32))
                idx += 1
    new_embeddings.insert(0, np.mean(new_embeddings, axis=0))
    new_word2idx["[PAD]"] = idx
    new_embeddings.append(np.zeros_like(new_embeddings[-1]))
    new_embeddings = torch.Tensor(np.asarray(new_embeddings))
    log.info(f"Reduced embeddings size from {len(embeddings)} to {len(new_embeddings)}")
    return new_embeddings, new_word2idx



import logging
from typing import Union

import numpy as np
import torch
from datasets.arrow_dataset import Dataset

from .al_strategy_utils import (
    assign_ue_scores_for_unlabeled_data,
    calculate_bald_score_cls,
    calculate_bald_score_ner,
    get_query_idx_for_selecting_by_number_of_tokens,
    get_X_pool_subsample
)
from .strategy_utils.batchbald.consistent_dropout import make_dropouts_consistent
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def entropy(probas):
    return - (probas * probas.log()).sum(-1)


def kl_div(p, q):
    return (p * (p / q).log()).sum()


def akim_batchbald(
    model,
    X_pool: Union[np.ndarray, Dataset, TransformersDataset],
    n_instances: int,
    **bald_kwargs,
):
    mc_iterations = bald_kwargs.get("mc_iterations", 10)
    use_stable_dropout = bald_kwargs.get("use_stable_dropout", False)
    use_subsample = bald_kwargs.get("use_subsample", False)

    lamb = bald_kwargs.get("lamb", None)
    if lamb is not None:
        label_name = model.data_config["label_name"]
        labels = bald_kwargs.get("X_train")[label_name]
        true_labels_probas = torch.Tensor(np.bincount(labels) / len(labels))

    if use_subsample:
        X_pool_subsample, subsample_indices = get_X_pool_subsample(
            X_pool, mc_iterations, model.seed
        )
    else:
        X_pool_subsample = X_pool

    # Make dropout consistent inside huggingface model
    if use_stable_dropout:
        make_dropouts_consistent(model.model)
    else:
        model.enable_dropout()

    if bald_kwargs.get("only_head_dropout", False):
        raise NotImplementedError
    else:
        # Stable dropout
        probas = []
        for _ in range(mc_iterations):
            if use_stable_dropout:
                # Reset masks
                model.enable_dropout()
                model.disable_dropout()
            probas_iter = model.predict_proba(
                X_pool_subsample, use_predict_loop=True, to_eval_mode=False
            )
            probas.append(probas_iter)

    query_idx = []
    unselected_idx = list(range(len(X_pool_subsample)))
    query_scores = []
    probas = torch.Tensor(probas).transpose(1, 0)
    av_probas = probas.mean(1)
    data_unc = entropy(probas).mean(-1)
    for i in range(n_instances):
        unc_iter = torch.empty(len(unselected_idx), dtype=torch.float32, device=av_probas.device)
        for j, idx in enumerate(unselected_idx):
            avg_query_probas = av_probas[query_idx + [idx]].mean(0)
            unc = entropy(avg_query_probas)
            if lamb is not None:
                unc -= lamb * kl_div(
                    true_labels_probas, avg_query_probas
                )
            unc_iter[j].copy_(unc)

        data_unc_iter = (data_unc[query_idx].sum(0) + data_unc[unselected_idx]) / (len(query_idx) + 1)
        scores_iter = unc_iter - data_unc_iter
        argmax = scores_iter.argmax().item()
        iter_query_idx = unselected_idx[argmax]
        query_idx.append(iter_query_idx)
        unselected_idx.remove(iter_query_idx)
        query_scores.append(scores_iter[argmax].item())

    uncertainty_estimates = np.empty(len(X_pool_subsample), dtype=np.float32)
    uncertainty_estimates[query_idx] = query_scores
    uncertainty_estimates[unselected_idx] = torch.cat((scores_iter[:argmax], scores_iter[argmax + 1:]))

    query = X_pool_subsample.select(query_idx)
    if use_subsample:
        # Update query_idx since now it corresponds to the sampled set
        query_idx = subsample_indices[query_idx]
        uncertainty_estimates = assign_ue_scores_for_unlabeled_data(
            len(X_pool), subsample_indices, uncertainty_estimates
        )

    return query_idx, query, uncertainty_estimates



from typing import Union, Tuple

from datasets import Dataset

from .load_arbitrary_dataset import (
    load_arbitrary_dataset_for_cls,
    load_conll_format_dataset_for_ner,
    load_from_csv,
)
from .load_from_url import load_data_from_url
from .load_from_json_or_csv import load_from_json_or_csv
from .load_huggingface_dataset import load_huggingface_dataset
from ..transformers_dataset import TransformersDataset


def load_data(
    config, task, cache_dir=None
) -> Tuple[
    Union[Dataset, TransformersDataset],
    Union[Dataset, TransformersDataset],
    Union[Dataset, TransformersDataset],
    Union[None, dict],
]:
    """
    :param config:
    :param task:
    :param cache_dir:
    :return: train_dataset, dev_dataset, test_dataset, id2label
    """
    if config.path == "url":
        return load_data_from_url(config, cache_dir)
    elif config.path != "datasets":
        if config.get("from_csv", None) is not None:
            return load_from_csv(config, task, cache_dir)
        elif task == "ner":
            return load_conll_format_dataset_for_ner(config)
        elif task in ["ats", "nmt"]:
            return load_from_json_or_csv(config, task, cache_dir)
        else:
            return load_arbitrary_dataset_for_cls(config)
    return load_huggingface_dataset(config, task, cache_dir)



from datasets.arrow_dataset import Dataset
import logging
from typing import Union
import numpy as np
import torch
from sklearn.neighbors import KNeighborsClassifier
from tqdm import tqdm

from .al_strategy_utils import _get_labels

from ..utils.get_embeddings import get_embeddings
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def cal(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **cal_kwargs,
):
    """
    Selects unlabeled data points from the pool,
    whose predictive likelihoods diverge the most from their neighbors in the training set.
    https://arxiv.org/pdf/2109.03764.pdf
    """
    num_neighbors = cal_kwargs.get("n_neighbors", 10)
    if num_neighbors > len(X_train):
        num_neighbors = len(X_train)
    logits = model.predict_logits(X_pool)
    train_logits = model.predict_logits(X_train)
    kwargs = dict(
        # Necessary
        model_wrapper=model,
        data_train=X_train,
        data_test=X_pool,
        # General
        data_is_tokenized=False,
        data_config=None,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        probas=logits,
        train_probas=train_logits,
        num_nei=num_neighbors,
    )

    uncertainty_estimates = calculate_cal_scores(**kwargs)
    query_idx = np.argpartition(uncertainty_estimates, -n_instances)[-n_instances:]

    query = X_pool.select(query_idx)

    return query_idx, query, uncertainty_estimates

# Contrastive Active Learning (CAL) https://aclanthology.org/2021.emnlp-main.51.pdf
def calculate_cal_scores(
    model_wrapper,
    data_train,
    data_test,
    probas,
    train_probas,
    use_activation: bool = False,
    use_spectralnorm: bool = False,
    data_is_tokenized=False,
    data_config=None,
    batch_size=100,
    to_numpy=True,
    num_nei=10,
    operator="mean",
):
    data_config = data_config if data_config is not None else model_wrapper.data_config
    kwargs = dict(
        # General
        model=model_wrapper.model,
        prepare_model=True,
        batch_size=batch_size,
        to_numpy=False,
        # DDU
        use_activation=use_activation,
        use_spectralnorm=use_spectralnorm,
        # Tokenization
        data_is_tokenized=data_is_tokenized,
        tokenizer=model_wrapper.tokenizer,
        task=model_wrapper.task,
        text_name=data_config["text_name"],
        label_name=data_config["label_name"],
    )

    train_embeddings = (
        get_embeddings(dataloader_or_data=data_train, **kwargs).detach().cpu()
    )
    test_embeddings = (
        get_embeddings(dataloader_or_data=data_test, **kwargs).detach().cpu()
    )

    if not isinstance(data_train, TransformersDataset):
        data_train = TransformersDataset(data_train)
    train_labels = _get_labels(data_train, data_config)

    distances = None
    num_adv = None

    neigh = KNeighborsClassifier(n_neighbors=num_nei)
    neigh.fit(X=train_embeddings, y=np.array(train_labels))

    criterion = torch.nn.KLDivLoss(reduction="none")

    kl_scores = []
    num_adv = 0
    distances = []
    pairs = []
    for unlab_i, candidate in enumerate(
        tqdm(
            zip(test_embeddings, probas),
            desc="Finding neighbours for every unlabeled data point",
        )
    ):
        # find indices of closesest "neighbours" in train set
        unlab_representation, unlab_logit = candidate
        distances_, neighbours = neigh.kneighbors(
            X=[candidate[0].numpy()], return_distance=True
        )
        distances.append(distances_[0])

        # remove outliers?
        # cur_mean_dist = np.max(distances_[0])

        labeled_neighbours_labels = train_labels[neighbours[0]]
        preds_neigh = [np.argmax(train_probas[n], axis=1) for n in neighbours]
        neigh_prob = torch.nn.functional.softmax(
            torch.Tensor(train_probas[neighbours]).to("cpu"), dim=-1
        )
        pred_candidate = [np.argmax(candidate[1])]

        uda_softmax_temp = 1
        candidate_log_prob = torch.nn.functional.log_softmax(
            torch.Tensor(candidate[1] / uda_softmax_temp).to("cpu"), dim=-1
        )
        kl = np.array(
            [
                torch.sum(criterion(candidate_log_prob, n), dim=-1).numpy()
                for n in neigh_prob
            ]
        )

        if operator == "mean":
            kl_scores.append(kl.mean())
        elif operator == "max":
            kl_scores.append(kl.max())
        elif operator == "median":
            kl_scores.append(np.median(kl))

    return np.array(kl_scores)



import numpy as np

from .utils_heads import (
    ElectraClassificationHeadCustom,
    ElectraClassificationHeadSN,
    ElectraNERHeadCustom,
    ElectraNERHeadSN,
    ElectraClassificationHS,
)


def is_custom_head(model):
    if not hasattr(model, "classifier"):
        return False
    if (
        isinstance(model.classifier, ElectraClassificationHeadCustom)
        or isinstance(model.classifier, ElectraNERHeadCustom)
        or isinstance(model.classifier, ElectraClassificationHeadSN)
        or isinstance(model.classifier, ElectraNERHeadSN)
        or isinstance(model.classifier, ElectraClassificationHS)
    ):
        return True
    return False


def unpad_features(features, labels):
    true_features = [f for (f, l) in zip(features, labels) if l != -100]
    true_labels = [l for (f, l) in zip(features, labels) if l != -100]

    return np.array(true_features), np.array(true_labels)


def pad_scores(unc, full_labels, flatten_labels):
    flatten_full_ue = np.zeros(shape=flatten_labels.shape)
    flatten_full_ue[flatten_labels != -100] = unc
    full_ue = flatten_full_ue.reshape(full_labels.shape)
    return full_ue



from typing import List


def align_labels_with_tokens(
    labels: List[int], word_ids: List[int], tag_for_subtokens=None
):
    """
    Function to adjust the labels according to the tokenized input
    :param labels: original list of tags
    :param word_ids: list with ids of the original words in the tokenized list of tokens;
    can be extracted via `inputs.word_ids()`, where `inputs = tokenizer(...)`
    :param tag_for_subtokens: how to tokenize subtokens (BPEs) of a word. Have three options:
    imagine we have a word `loving` -> [`lov`, `###ing`].
    Option 1 (default) = None: `lov` -> B-TAG, `###ing` -> O
    Option 2 = 'i': `lov` -> B-TAG, `###ing` -> I-TAG
    Option 3 = 'b': `lov` -> B-TAG, `###ing` -> B-TAG
    """
    assert tag_for_subtokens in [
        None,
        "i",
        "b",
    ], "Param :tag_for_subtokens must be one of `None`, `'i'`, `'b'`!"
    new_labels = []
    current_word = None
    for word_id in word_ids:
        if word_id is None:
            # Special token
            new_labels.append(-100)
        elif word_id != current_word:
            # Start of a new word!
            current_word = word_id
            new_labels.append(labels[word_id])
        else:
            # Same word as previous token
            if tag_for_subtokens is not None:
                label = labels[word_id]
                # If the label is B-XXX we change it to I-XXX
                if tag_for_subtokens == "i" and label % 2 == 1:
                    label += 1
            else:
                label = -100
            new_labels.append(label)

    return new_labels



from collections import defaultdict
from pathlib import Path
from typing import Union, List

import numpy as np
from .utils.general import pickle_load
from tqdm import tqdm


class ActiveSelector:
    def __init__(
        self,
        scores_dir: Path or str,
        idxs: Union[int, List[int], np.ndarray[int]],
        ngramms_range: tuple = (1, 2),
        query_ids: list or tuple = (),
        strategy: str = "mmr_rec_prec",  # "mmr_f1" or "mmr_rec_prec"
        lamb: float = 0.5,
    ):
        self.scores_dir = Path(scores_dir)
        self.ngramms_range = ngramms_range
        self.query_ids = list(query_ids)

        if isinstance(idxs, int):
            self.idxs = np.arange(idxs)
        else:
            self.idxs = np.array(idxs)

        self.strategy = strategy
        self.lamb = lamb
        self.cache = defaultdict(list)

    def query(self, num_queries=1):
        for i_query in range(num_queries):
            self._query_one_sample()
        return self.query_ids

    def _query_one_sample(self):
        ids_in = self.query_ids
        ids_out = np.setdiff1d(self.idxs, ids_in)
        length_out_minus_one = len(ids_out) - 1
        cache = self.cache
        strategy = self.strategy

        if len(cache) == 0:
            for idx in tqdm(ids_out):
                for ngramm in self.ngramms_range:
                    if strategy == "mmr_f1":
                        scores_in_idx = scores_out_idx = pickle_load(
                            self.scores_dir / f"fscores_{ngramm}_{idx}"
                        )
                    elif strategy == "mmr_rec_prec":
                        # These are "how much info of each text is contained in text `idx`"
                        # i.e. [precision(text_0, text_idx), precision(text_1, text_idx), ...]
                        scores_in_idx = pickle_load(
                            self.scores_dir / f"recalls_{ngramm}_{idx}"
                        )
                        # These are "how much info of text `idx` is contained in each text "
                        # i.e. [recall(text_0, text_idx), recall(text_1, text_idx), ...]
                        scores_out_idx = pickle_load(
                            self.scores_dir / f"precisions_{ngramm}_{idx}"
                        )
                    else:
                        raise NotImplementedError

                    sim_in = scores_in_idx[ids_in].max() if len(ids_in) else 0
                    sim_out = (scores_out_idx[ids_out].sum() - 1) / length_out_minus_one
                    cache[f"idx_out_scores_{ngramm}"].append(sim_out)
                    cache[f"idx_in_scores_{ngramm}"].append(sim_in)

            for ngramm in self.ngramms_range:
                cache[f"idx_out_scores_{ngramm}"] = np.array(
                    cache[f"idx_out_scores_{ngramm}"], dtype=float
                )
                cache[f"idx_in_scores_{ngramm}"] = np.array(
                    cache[f"idx_in_scores_{ngramm}"], dtype=float
                )
        else:
            n = len(ids_out)
            for ngramm in self.ngramms_range:
                last_query_out_scores = cache[f"query_out_out_scores_{ngramm}"]
                last_query_in_scores = cache[f"query_out_in_scores_{ngramm}"]
                # n instead if (n+1) since we ecxlude the sim of each item with itself
                cache[f"idx_out_scores_{ngramm}"] = (
                    cache[f"idx_out_scores_{ngramm}"] * n - last_query_out_scores
                ) / (n - 1)
                cache[f"idx_in_scores_{ngramm}"] = np.maximum(
                    cache[f"idx_in_scores_{ngramm}"], last_query_in_scores
                )

        ngramm_scores = []
        for ngramm in self.ngramms_range:
            cache[f"idx_scores_{ngramm}"] = (
                self.lamb * cache[f"idx_out_scores_{ngramm}"]
                - (1 - self.lamb) * cache[f"idx_in_scores_{ngramm}"]
            )
            ngramm_scores.append(cache[f"idx_scores_{ngramm}"])

        ngramm_scores = cache["ngramm_scores"] = np.stack(ngramm_scores, axis=1)
        scores = np.prod(np.maximum(ngramm_scores, 0), axis=1) ** (
            1 / len(self.ngramms_range)
        )
        one_of_values_lower_zero = (ngramm_scores < 0).sum(axis=1) > 0
        modified_scores = (
            np.prod(np.maximum(ngramm_scores + 1, 0), axis=1)
            ** (1 / len(self.ngramms_range))
            - 1
        )
        cache["scores"] = final_scores = (
            scores * (~one_of_values_lower_zero)
            + modified_scores * one_of_values_lower_zero
        )

        argmax = np.argmax(final_scores)
        id_query = int(ids_out[argmax])  # to cast it to standard int type
        self.query_ids.append(id_query)
        # Cache query out scores
        ids_out = np.setdiff1d(ids_out, id_query)
        for ngramm in self.ngramms_range:
            if strategy == "mmr_f1":
                scores_in_idx = scores_out_idx = pickle_load(
                    self.scores_dir / f"fscores_{ngramm}_{id_query}"
                )
            elif strategy == "mmr_rec_prec":
                # Note that we use recall for `out` and precision for `in` (opposite to what was done earlier)
                # since precision(x1, x2) = recall(x2, x1)
                scores_in_idx = pickle_load(
                    self.scores_dir / f"precisions_{ngramm}_{id_query}"
                )
                scores_out_idx = pickle_load(
                    self.scores_dir / f"recalls_{ngramm}_{id_query}"
                )
            else:
                raise NotImplementedError
            # First "out" refers to the fact that these are scores with "out" samples; second - to the metric ("in" / "out")
            cache[f"query_out_out_scores_{ngramm}"] = scores_out_idx[ids_out].ravel()
            cache[f"query_out_in_scores_{ngramm}"] = scores_in_idx[ids_out].ravel()
            # Update cached in and out scores: remove the score of the query
            cache[f"idx_out_scores_{ngramm}"] = np.delete(
                cache[f"idx_out_scores_{ngramm}"], argmax
            )
            cache[f"idx_in_scores_{ngramm}"] = np.delete(
                cache[f"idx_in_scores_{ngramm}"], argmax
            )

        self.cache = cache



import sys
from importlib import import_module

from transformers import set_seed

from ..transformers.wrapper_encoder import WrapperEncoder
from ...models import PYTORCH_INIT_MODELS_DICT


class PytorchEncoderWrapper(WrapperEncoder):
    def __init__(self, embeddings, word2idx, **base_kwargs):
        self.embeddings = embeddings
        self.word2idx = word2idx
        super().__init__(**base_kwargs)

    def get_model_and_tokenizer(
        self,
        model_cfg,
        id2label: dict = None,
        seed: int = 42,
        cache_dir=None,
        embeddings=None,
        word2idx=None,
    ):
        set_seed(seed)
        num_labels = model_cfg.num_labels if id2label is None else len(id2label)
        # build CNN for text classification
        init_function = self.get_init_function()
        model, hf_tokenizer = init_function(model_cfg, embeddings, word2idx, num_labels)
        return model, hf_tokenizer

    def get_init_function(self):
        if self.model_config.checkpoint in PYTORCH_INIT_MODELS_DICT.keys():
            # try to get function from implemented in framework functions
            init_function = PYTORCH_INIT_MODELS_DICT[self.model_config.checkpoint][
                "model"
            ]
        else:
            # use a custom function and model
            _, init_function = self.get_model_from_path()
        return init_function

    def get_init_model_kwargs(self):
        return {
            "id2label": self.id2label,
            "embeddings": self.embeddings,
            "word2idx": self.word2idx,
        }

    def set_tokenizer(self):
        init_function = self.get_init_function()
        try:
            num_labels = self.model_config.num_labels
        except:
            # we doesn't really need this param for tokenizer
            num_labels = 2
        _, hf_tokenizer = init_function(
            self.model_config,
            self.embeddings,
            self.word2idx,
            num_labels,
            only_tokenizer=True,
        )
        self.tokenizer = hf_tokenizer

    def get_model_from_path(self):
        model_path = self.model_config.checkpoint
        if (not model_path.endswith("/")) and ("/" in model_path):
            path = model_path[: model_path.rindex("/")]
        elif "/" in model_path:
            path = model_path[: model_path[:-2].rindex("/")]
        else:
            path = "./"
        model_filename = model_path.split("/")[-1] or model_path.split("/")[-2]
        if model_filename.endswith(".py"):
            model_filename = model_filename[:-3]
        sys.path.append(path)

        models_dict = getattr(import_module(model_filename), "MODELS_DICT", None)
        model_class = getattr(import_module(model_filename), models_dict["model"], None)
        model_constructor = getattr(
            import_module(model_filename), models_dict["constructor"], None
        )
        return model_class, model_constructor



import logging
from typing import Union

import numpy as np
from datasets.arrow_dataset import Dataset
from sklearn.metrics import pairwise_distances

from .al_strategy_utils import take_idx
from ..utils.get_embeddings import get_embeddings
from ..utils.transformers_dataset import TransformersDataset

log = logging.getLogger()


def furthest_first(X, X_set, n):
    m = np.shape(X)[0]
    if np.shape(X_set)[0] == 0:
        min_dist = np.tile(float("inf"), m)
    else:
        dist_ctr = pairwise_distances(X, X_set)
        min_dist = np.amin(dist_ctr, axis=1)

    idxs = []

    for i in range(n):
        idx = min_dist.argmax()
        idxs.append(idx)
        dist_new_ctr = pairwise_distances(X, X[[idx], :])
        for j in range(m):
            min_dist[j] = min(min_dist[j], dist_new_ctr[j, 0])

    return np.array(idxs)


def coreset(
    model,
    X_pool: Union[Dataset, TransformersDataset],
    n_instances: int,
    X_train: Union[Dataset, TransformersDataset],
    **coreset_kwargs,
):
    kwargs = dict(
        # General
        prepare_model=True,
        # use_activation=use_activation,
        # use_spectralnorm=use_spectralnorm,
        data_is_tokenized=False,
        batch_size=model._batch_size_kwargs.eval_batch_size,
        to_numpy=True,
        # Tokenization
        tokenizer=model.tokenizer,
        task=model.task,
        text_name=model.data_config["text_name"],
        label_name=model.data_config["label_name"],
    )

    train_features = get_embeddings(model.model, X_train, **kwargs)
    test_features = get_embeddings(model.model, X_pool, **kwargs)

    query_idx = furthest_first(test_features, train_features, n_instances)

    query = X_pool.select(query_idx)

    # Uncertainty estimates are not defined for CoreSet
    uncertainty_estimates = np.zeros(len(X_pool))

    return query_idx, query, uncertainty_estimates
